from loguru import logger
from shutil import rmtree, copytree
from uuid import UUID
import json
import pprint
import typing

from handshake.services.DBService.lifecycle import (
    init_tortoise_orm,
    close_connection,
    set_default_config,
    config_file,
    READ_ONLY,
    ALLOW_WRITE,
)
from pathlib import Path
from typing import Dict
from handshake.services.DBService.models.config_base import ConfigKeys, ConfigBase
from tortoise.expressions import Q


def delete_test_attachment(run: UUID, attachmentFolder: Path):
    entry = attachmentFolder / str(run)
    if not entry.exists():
        return

    logger.warning("Deleting the attachments for run {}", run)
    rmtree(entry)


def dump_config(path: Path, to_dump: typing.List):
    formatted_dump: typing.Dict[str, str] = dict(to_dump)

    # not required
    for key in READ_ONLY:
        if key in formatted_dump:
            formatted_dump.pop(key)

    # show only required
    config_file(path).write_text(json.dumps(formatted_dump, indent=4))


async def setConfig(path: Path, feed: Dict[ConfigKeys, str], set_default: bool):
    await init_tortoise_orm(path)

    if set_default:
        await set_default_config(path)
    if feed:
        to_change = await ConfigBase.filter(Q(key__in=feed.keys())).all()
        for to in to_change:
            if to.key not in ALLOW_WRITE:
                continue

            to.value = feed[to.key]
            await ConfigBase.bulk_update(to_change, fields=["value"])

    to_dump = await ConfigBase.all().values_list()
    dump_config(path, to_dump)

    pprint.pprint(to_dump, indent=4)
    await close_connection()



from typing import Optional
import traceback
from handshake.services.DBService.models.result_base import (
    SuiteBase,
    RollupBase,
    RetriedBase,
    SessionBase,
)
from handshake.services.DBService.models.config_base import TestConfigBase
from handshake.services.DBService.models.dynamic_base import TaskBase
from handshake.services.DBService.models.enums import Status, SuiteType
from tortoise.expressions import Q
from tortoise.functions import Count, Lower, Sum
from loguru import logger
from handshake.services.SchedularService.register import skip_test_run
from itertools import chain
from asyncio import gather


def fetch_key_from_status(passed, failed, skipped):
    return (
        Status.FAILED
        if failed > 0
        else Status.PASSED
        if passed > 0 or skipped == 0
        else Status.SKIPPED
    )


class PatchTestSuite:
    def __init__(self, suite_id, test_id):
        self.suite: Optional[SuiteBase] = None
        self.related_task: Optional[TaskBase] = None
        self.test_config: Optional[TestConfigBase] = None
        self.suite_id = str(suite_id)
        self.test_id = str(test_id)

    async def fetch_records(self):
        self.suite, self.related_task, self.test_config = await gather(
            SuiteBase.filter(suiteID=self.suite_id).first(),
            TaskBase.filter(ticketID=self.suite_id).first(),
            TestConfigBase.filter(test_id=self.test_id).first(),
        )

    async def do_we_need_to_patch(self):
        logger.info("Patching Suite: {} | {}", self.suite.suiteID, self.suite.title)

        if self.suite.standing != Status.YET_TO_CALCULATE:
            logger.warning("Skipping already patch suite for: {}", self.suite.suiteID)
            await self.mark_processed()
            return False

        pending_child_tasks = await SuiteBase.filter(
            Q(parent=self.suite.suiteID)
            & (Q(standing=Status.PENDING) | Q(standing=Status.YET_TO_CALCULATE))
        ).exists()

        if pending_child_tasks:
            logger.warning(
                "There are some child suites, which are not yet processed,"
                " so will process {} suite in the next iteration",
                self.suite.suiteID,
            )
            await self.pick_it_later()
            return False
        return True

    async def pick_it_later(self):
        self.related_task.picked = False
        await self.related_task.save()

    async def mark_processed(self):
        self.related_task.processed = True
        await self.related_task.save()

    """
    returns true if suite is patched else false
    """

    async def patch_suite(self) -> bool:
        await self.fetch_records()
        if not await self.do_we_need_to_patch():
            return False

        await gather(
            self.patch_status(),
            self.patch_rollup_value_for_errors(),
            self.patch_rollup_table(),
            self.patch_retried_records(),
        )

        logger.info("Successfully processed suite: {}", self.suite.suiteID)
        await self.mark_processed()
        return True

    async def patch_status(self):
        entities = SuiteBase.filter(parent=self.suite.suiteID)

        results = dict(
            await (
                entities.annotate(count=Count("standing"), status=Lower("standing"))
                .group_by("standing")
                .values_list("status", "count")
            )
        )
        results["standing"] = fetch_key_from_status(
            results.get("passed", 0),
            results.get("failed", 0),
            results.get("skipped", 0),
        )
        results["tests"] = await entities.count()

        await self.suite.update_from_dict(results)
        await self.suite.save()

    async def patch_rollup_value_for_errors(self):
        errors = list(
            chain.from_iterable(
                map(
                    lambda suite_with_errors: map(
                        lambda error: error
                        | dict(
                            mailedFrom=error.get("mailedFrom", [])
                            + [str(suite_with_errors.suiteID)]
                        ),
                        suite_with_errors.errors,
                    ),
                    await (
                        SuiteBase.filter(Q(parent=self.suite.suiteID) & ~Q(errors="[]"))
                        .only("errors", "suiteID")
                        .all()
                    ),
                )
            )
        )

        await self.suite.update_from_dict(
            dict(
                errors=errors,
            )
        )
        await self.suite.save()

    async def patch_rollup_table(self):
        required = ("passed", "failed", "skipped", "tests")

        direct_entities = (
            await (
                SuiteBase.filter(parent=self.suite_id, suiteType=SuiteType.TEST)
                .only("parent", *required)
                .group_by("parent")
                .annotate(**{key: Sum(key) for key in required})
                .first()
                .values(*required)
            )
            or {}
        )

        suites = await SuiteBase.filter(
            suiteType=SuiteType.SUITE, parent=self.suite_id
        ).values_list("suiteID", flat=True)

        indirect_entities = (
            (
                await RollupBase.filter(Q(suite_id__in=suites))
                .annotate(**{key: Sum(key) for key in required})
                .first()
                .values(*required)
            )
            if suites
            else None
        ) or {}

        await RollupBase.create(
            suite_id=self.suite_id,
            **{
                key: direct_entities.get(key, 0) + indirect_entities.get(key, 0)
                for key in required
            },
        )

    async def patch_retried_records(self):
        if not (self.test_config and self.test_config.fileRetries > 0):
            return

        # assuming we would need to match the retries
        suite = await SuiteBase.filter(suiteID=self.suite_id).first()
        if suite.retried == 0:
            return await RetriedBase.create(
                tests=[self.suite_id], suite_id=self.suite_id, length=1
            )

        matching_previous_suites = (
            await SuiteBase.filter(
                Q(title=suite.title)
                & Q(session__test_id=self.test_id)
                & Q(file=suite.file)
                & Q(tags=suite.tags)
                & Q(retried=suite.retried - 1)
                & Q(suiteType=suite.suiteType)
                & ~Q(suiteID=suite.suiteID)
                & Q(ended__lte=suite.started)
            )
            .order_by("started")
            .values_list("suiteID", flat=True)
        )

        previous = (
            await RetriedBase.filter(
                Q(length=suite.retried) & Q(suite_id__in=matching_previous_suites)
            )
            .order_by("modified")
            .first()
        )

        previous_suite = await SuiteBase.filter(suiteID=previous.suite_id).first()
        await previous_suite.update_from_dict(dict(standing=Status.RETRIED))
        await previous_suite.save()

        session = await SessionBase.filter(
            sessionID=(await previous_suite.session).sessionID
        ).first()
        await session.update_from_dict(dict(retried=True))
        await session.save()

        await previous.update_from_dict(
            dict(
                length=previous.length + 1,
                tests=previous.tests + [self.suite_id],
                suite_id=suite.suiteID,
            )
        )

        await previous.save()
        return previous

    async def fall_back(self):
        await skip_test_run(
            self.test_id,
            f"Failed to patch the test suite, found an error in calculation: {traceback.format_exc()}",
            suiteID=self.suite_id,
        )


async def patchTestSuite(suiteID: str, testID: str):
    patcher = PatchTestSuite(suiteID, testID)
    try:
        has_patched = await patcher.patch_suite()
        return has_patched
    except Exception:
        logger.exception("Failed to patch test suite")
        await patcher.fall_back()
        return False



from handshake.services.Endpoints.internalEndpoints import one_liners
from handshake.services.DBService.shared import APP_NAME
from handshake.services.Endpoints.blueprints.coreEndpoints import service
from handshake.services.Endpoints.blueprints.writeServices import writeServices
from handshake.services.Endpoints.errorHandling import handle_validation_error
from sanic import Sanic
from pydantic import ValidationError

service_provider = Sanic(APP_NAME)
service_provider.config.TOUCHUP = False
service_provider.blueprint(one_liners)
service_provider.blueprint(service)
service_provider.blueprint(writeServices)

service_provider.error_handler.add(ValidationError, handle_validation_error)



from tortoise.models import Model
from tortoise.fields import (
    DatetimeField,
    IntField,
    FloatField,
    JSONField,
    CharEnumField,
    UUIDField,
    CharField,
    ReverseRelation,
    ForeignKeyField,
    BooleanField,
    ForeignKeyRelation,
    TextField,
)
from handshake.services.DBService.models.enums import Status, SuiteType, LogType


class CommandReportFields(Model):
    started = DatetimeField(descrption="Start Datetime field")
    ended = DatetimeField(null=True, description="End Datetime field")
    tests = IntField(default=0, null=False, description="No. of the test entries")
    passed = IntField(default=0, null=False, description="Passed test entities")
    failed = IntField(default=0, null=False, description="Failed test entities")
    skipped = IntField(
        default=0, null=False, description="test entities that were skipped"
    )
    duration = FloatField(default=0, null=False, description="duration of test entity")
    retried = IntField(
        default=0, null=False, description="number of the retries performed"
    )

    class Meta:
        abstract = True


class CommonDetailedFields(CommandReportFields):
    class Meta:
        abstract = True


class EntityBaseSpecific:
    standing = CharEnumField(
        Status, description="status of the test run", default=Status.PENDING
    )

    class Meta:
        abstract = True


class RunBase(CommonDetailedFields, EntityBaseSpecific):
    table = "RunBase"
    testID = UUIDField(pk=True)
    sessions = ReverseRelation["SessionBase"]
    tasks = ReverseRelation["TaskBase"]
    config = ReverseRelation["TestConfigBase"]
    started = DatetimeField(null=False, auto_now_add=True)
    projectName = CharField(
        max_length=30, null=False, description="Name of the project"
    )
    specStructure = JSONField(
        description="file structure of spec files", default=dict()
    )
    suiteSummary = JSONField(
        description="summary of the suites",
        default=dict(count=0, passed=0, failed=0, skipped=0),
    )
    exitCode = IntField(
        null=False, default=0, description="Exit code for the test execution"
    )


class SessionBase(CommonDetailedFields):
    table = "SessionBase"
    test: ForeignKeyRelation[RunBase] = ForeignKeyField(
        "models.RunBase", related_name="sessions", to_field="testID"
    )
    retried = BooleanField(
        description="was this session omitted, because there would be one more retried session",
        default=False,
        required=False,
    )
    suites = ReverseRelation["SuiteBase"]
    sessionID = UUIDField(pk=True)
    entityName = CharField(max_length=30, default="")
    entityVersion = CharField(max_length=20, default="")
    simplified = TextField(
        default="", description="browser name & version &/ platform name included"
    )
    hooks = IntField(default=0, null=False, description="Number of hooks used")


class SuiteBase(EntityBaseSpecific, CommandReportFields):
    table = "SuiteBase"
    # https://tortoise.github.io/models.html#the-db-backing-field
    # so we require session_id instead of sessionID
    session: ForeignKeyRelation[SessionBase] = ForeignKeyField(
        "models.SessionBase", related_name="suites", to_field="sessionID"
    )
    attachments = ReverseRelation["AttachmentBase"]
    retries = ReverseRelation["RetriedBase"]
    rolled_up = ReverseRelation["RollupBase"]

    suiteID = UUIDField(pk=True)
    suiteType = CharEnumField(
        SuiteType,
        description="Specifies whether if it is a test suite or test case",
        null=False,
    )
    title = TextField(max_length=225)
    file = TextField(max_length=150, null=False, description="path to the spec file")
    description = TextField(
        null=True, default="", description="Summary if provided for the test entity"
    )
    parent = CharField(max_length=45, description="Parent Suite's ID", default="")
    tags = JSONField(
        description="comma separated list of tags used by the framework to filter the suites or spec files",
        default=[],
    )
    modified = DatetimeField(
        auto_now=True, description="Modified timestamp", null=False
    )
    errors = JSONField(description="Errors found", default=[], null=True)


class RollupBase(Model):
    tests = IntField(default=0, null=False, description="Rolled up Test Entities")
    passed = IntField(
        default=0, null=False, description="Rolled up Passed test entities"
    )
    failed = IntField(
        default=0, null=False, description="Rolled up Failed test entities"
    )
    skipped = IntField(
        default=0, null=False, description="Rolled up test entities that were skipped"
    )
    suite: ForeignKeyRelation[SuiteBase] = ForeignKeyField(
        "models.SuiteBase", related_name="rollup", to_field="suiteID"
    )


class RetriedBase(Model):
    tests = JSONField(
        default=[],
        description="details of all suites which were retried for a particular test case",
    )
    suite: ForeignKeyRelation[SuiteBase] = ForeignKeyField(
        "models.SuiteBase", related_name="retries", to_field="suiteID"
    )

    length = IntField(
        description="Just the length of the tests cols", default=1, null=False
    )

    modified = DatetimeField(
        auto_now=True, description="Modified timestamp", null=False
    )


class TestLogBase(Model):
    test: ForeignKeyRelation[RunBase] = ForeignKeyField(
        "models.RunBase", related_name="log", to_field="testID", pk=True
    )
    type = CharEnumField(LogType, description="Log type", null=False)
    message = TextField(description="Log Message", null=False)
    feed = JSONField(
        description="If you want to share any feed here", null=True, default={}
    )
    dropped = DatetimeField(auto_now=True, description="timestamp", null=False)



import datetime
import pprint
from sqlite3 import connect, sqlite_version_info
from click import (
    group,
    argument,
    secho,
    version_option,
    pass_context,
    Context,
    Path as C_Path,
)
from handshake import __version__
from handshake.services.DBService.migrator import (
    check_version,
    migration,
    DB_VERSION,
    MigrationTrigger,
)
from handshake.services.SchedularService.start import Scheduler
from handshake.services.SchedularService.handleTestResults import (
    setConfig,
)
from loguru import logger
import json
from handshake.services.DBService.lifecycle import config_file, close_connection
from click import option
from pathlib import Path
from handshake.services.DBService.shared import db_path
from handshake.services.DBService.models.config_base import ConfigKeys
from tortoise import run_async
from asyncio import run


@group(
    name="Handshake",
    short_help="Handshake command",
    help=f"""

{'{:*^69}'.format(" Welcome to Handshake's CLI ")}

Handshake simplifies the collection and processing of your test results. The Handshake CLI helps you to interact with 
the stored results or with the server. so for each command it requires a <path> argument, representing the collection 
folder, a place where you could have the database or were planning to store the results [not the HTML files].

[ROOT-DIR] >> [COLLECTION_NAME] (*we need this) >> TeStReSuLtS.db & [Attachments] (for getting this).

{'{:*^69}'.format(" Glossary ")}
""",
)
@version_option(f"{__version__}, DB: {DB_VERSION}")
def handle_cli():
    pass


general_requirement = argument(
    "COLLECTION_PATH", nargs=1, type=C_Path(exists=True, dir_okay=True), required=True
)
general_but_optional_requirement = argument(
    "COLLECTION_PATH", nargs=1, type=C_Path(dir_okay=True), required=True
)
observed_version = option(
    "--version",
    "-v",
    default="",
    type=str,
    required=False,
    show_default=True,
    help="Used Internally with the reporters,"
    " reporters will pass the version of the expected handshake server if it matches,"
    " we run the server else we terminate the execution.",
)


def break_if_mismatch(expected: str) -> bool:
    if expected:
        assert expected.strip() == __version__, (
            f"Mismatch between expected version: {expected} "
            f"and the actual version v{__version__}"
        )
    return True


@general_requirement
@handle_cli.command()
def db_version(collection_path):
    return check_version(path=db_path(collection_path))


@handle_cli.command(
    short_help="checks the version of the sqlite3 installed in your system"
)
def check_sqlite():
    assert int(sqlite_version_info[0]) >= 3, "Required version is >= 3."
    assert (
        int(sqlite_version_info[1]) >= 38
    ), "Required Version is >= 3.38,  for supporting our sql scripts, for supporting our sql scripts"
    logger.info("your sqlite version is fine as per our need")


@general_requirement
@handle_cli.command(
    short_help="Migrates the database to the latest version as per the handshake executable.",
    help="it's a command to execute the required migration scripts, note; this command would be executed "
    "automatically whenever we run patch or run-app command",
)
def migrate(collection_path: str):
    return migration(db_path(collection_path), MigrationTrigger.CLI)


@handle_cli.command(
    short_help="Processes the collected results and even could export the test results",
    help="runs an async loop, schedules some tasks to patch some your test results "
    "so you can see it in the way we need. you can pass the output directory to generate the report",
)
@option(
    "--log-file",
    "-l",
    help="give me file name to store the logs for the patch command.",
    type=C_Path(),
    default="",
    required=False,
)
@option(
    "-r",
    "--reset",
    default=False,
    show_default=True,
    help="re-calculates the values for the test runs",
    type=bool,
    is_flag=True,
)
@general_requirement
# TODO: Support the max runs feature, allows user to only export certain number of test runs
# @option(
#     "-mr",
#     "--max-runs",
#     type=int,
#     default=100,
#     help="Asks Sanic to set the use max. number of workers",
# )
@option(
    "--build",
    "-b",
    required=False,
    help="builds the dashboard output at the build output folder",
    type=C_Path(exists=True, file_okay=True, readable=True),
)
@option(
    "--include",
    "-i",
    required=False,
    help="generates the Import Data folder inside the test results (used for internal purposes only)",
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@option("--out", "-o", type=C_Path(dir_okay=True, writable=True), required=False)
def patch(
    collection_path,
    log_file: str,
    reset: bool = False,
    build: str = None,
    out: str = None,
    include=False,
):
    if log_file:
        logger.add(
            log_file if log_file.endswith(".log") else f"{log_file}.log",
            backtrace=True,
            diagnose=True,
        )

    if not Path(collection_path).is_dir():
        raise NotADirectoryError(collection_path)

    scheduler = Scheduler(collection_path, out, reset, build, include)
    try:
        run(scheduler.start())
    except (KeyboardInterrupt, SystemExit):
        logger.warning("Scheduler terminated explicitly...")
        run(close_connection())


@handle_cli.command(
    help="returns the version of the handshake", short_help="example: 1.0.0"
)
def v():
    secho(__version__)


@handle_cli.command(
    short_help="Quick Config for test collection",
    help="configures few values which is used while processing your reports., ignore the options if not required for "
    "update",
)
@general_but_optional_requirement
@option(
    "--max_runs",
    "-mr",
    default=-1,
    help="Max. Number of runs to keep. NOTE: should be >1",
)
def config(collection_path, max_runs):
    saved_db_path = db_path(collection_path)

    set_default_first = not Path(collection_path).exists()
    if set_default_first:
        Path(collection_path).mkdir(exist_ok=True)

    feed_from = config_file(saved_db_path)
    feed = dict() if not feed_from.exists() else json.loads(feed_from.read_text())
    if max_runs > 1:
        feed[ConfigKeys.maxRunsPerProject] = max_runs

    run_async(setConfig(saved_db_path, feed, set_default_first))


@handle_cli.group(
    name="db",
    short_help="Commands that will/(try to) fetch mostly asked info. from db",
    help="you can query using db following the subcommands, which are created to provide mostly asked info. from db",
)
@version_option(DB_VERSION, prog_name="handshake-db")
@general_requirement
def db(collection_path: str):
    if not Path(db_path(collection_path)).exists():
        raise FileNotFoundError(
            f"db path was not found in the collections: {collection_path}"
        )


@db.command(short_help="fetches the timestamp of the latest run")
@option(
    "-p",
    "--allow-pending",
    default=False,
    show_default=True,
    help="consider runs, whose status are pending",
    type=bool,
    is_flag=True,
)
@pass_context
def latest_run(ctx: Context, allow_pending: bool):
    db_file = db_path(Path(ctx.parent.params["collection_path"]))
    pipe = connect(db_file)
    result = pipe.execute(
        "SELECT PROJECTNAME, ENDED FROM RUNBASE WHERE ENDED <> '' ORDER BY STARTED LIMIT 1"
        if not allow_pending
        else "SELECT PROJECTNAME, ENDED FROM RUNBASE ORDER BY STARTED LIMIT 1"
    ).fetchone()

    secho(
        "No Test Runs were found"
        if not result
        else (
            result[0],
            datetime.datetime.fromisoformat(result[1]).astimezone().strftime("%c %Z"),
        ),
        fg="bright_yellow" if not result else "bright_magenta",
    )

    pipe.close()


@db.command(
    short_help="fetches the number of yet to patch task",
    help="returns list of tasks of form: (ticket_id, task_type, dropped_date, is_picked, test_id)",
)
@pass_context
def yet_to_process(ctx: Context):
    db_file = db_path(Path(ctx.parent.params["collection_path"]))
    pipe = connect(db_file)
    result = pipe.execute(
        "SELECT ticketID, type, STRFTIME('%d/%m/%Y, %H:%M', dropped, 'localtime'), picked, test_id FROM TASKBASE WHERE "
        "PROCESSED = 0"
    ).fetchall()

    secho(
        "No Pending Tasks"
        if not result
        else f"pending tasks:\n {pprint.pformat(result)}",
        fg="bright_green" if not result else "bright_yellow",
    )

    pipe.close()



from handshake.services.DBService.shared import get_test_id
from handshake.services.DBService.lifecycle import close_connection
from sanic.blueprints import Blueprint
from sanic.response import HTTPResponse, text
from sanic.request import Request
from handshake.services.SchedularService.register import register_patch_test_run

one_liners = Blueprint(name="one_liners", url_prefix="/")


# Here we will have requests which are very simple but used at crucial points


@one_liners.get("/")
async def health_status(request: Request) -> HTTPResponse:
    return text("1")


@one_liners.put("/done")
async def set_done(request: Request) -> HTTPResponse:
    task = await register_patch_test_run(get_test_id())
    return text(task.ticketID)


# bye is core request, so make sure to handle it carefully
@one_liners.post("/bye")
async def bye(request: Request) -> HTTPResponse:
    await close_connection()
    request.app.m.terminate()

    return text("1")



from handshake.services.DBService.models.result_base import TestLogBase, LogType
from handshake.services.DBService.models.dynamic_base import TaskBase
from loguru import logger
from typing import Optional
from tortoise.expressions import Q
from typing import Union
from uuid import UUID


async def skip_test_run(test_id: Union[str, UUID], reason: str, **extra) -> False:
    logger.error(reason)
    await TestLogBase.create(
        test_id=str(test_id),
        feed=extra,
        type=LogType.ERROR,
        message=reason,
    )
    return False


async def pruneTasks(task_id: Optional[str] = ""):
    if task_id:
        logger.error("Deleting Few Tasks as per request")
    else:
        logger.warning("Pruning some Tasks, which are related to error test runs")

    to_prune = await TaskBase.filter(
        Q(
            test_id__in=await TestLogBase.filter(type=LogType.ERROR)
            .filter(
                test_id__in=await TaskBase.filter(processed=False)
                .only("test_id")
                .distinct()
                .values_list("test_id", flat=True)
            )
            .only("test_id")
            .distinct()
            .values_list("test_id", flat=True)
        )
    )

    pruned = []
    for task in to_prune:
        logger.error(
            "Found an error in this task: {}. hence marking it as processed."
            " Please report it as an issue if it was not expected.",
            task.ticketID,
        )
        task.processed = True
        task.picked = True
        pruned.append(task)

    await TaskBase.bulk_update(
        pruned,
        (
            "picked",
            "processed",
        ),
        100,
    )

    if not to_prune:
        logger.debug("No Tasks were pruned.")

    if task_id:
        # removing the prune task
        await TaskBase.filter(ticketID=task_id).delete()



from handshake.services.DBService.models.types import AttachmentType
from handshake.services.DBService.models.result_base import SuiteBase
from tortoise.models import Model
from tortoise.fields import (
    JSONField,
    CharEnumField,
    ForeignKeyField,
    ForeignKeyRelation,
    TextField,
    UUIDField,
)


class AttachmentFields(Model):
    attachmentValue = JSONField(
        description="An attachment value", default={"value": ""}
    )
    description = TextField(
        null=False, description="Description Field for the attachment"
    )
    type = CharEnumField(
        AttachmentType,
        description="Type of an attachment, refer the enums to get an idea",
    )

    class Meta:
        abstract = True


class AttachmentBase(AttachmentFields):
    table = "AttachmentBase"
    entity: ForeignKeyRelation[SuiteBase] = ForeignKeyField(
        "models.SuiteBase", related_name="attachment", to_field="suiteID"
    )


class StaticBase(AttachmentFields):
    attachmentID = UUIDField(pk=True)
    entity: ForeignKeyRelation[SuiteBase] = ForeignKeyField(
        "models.SuiteBase", related_name="staticAttachments", to_field="suiteID"
    )



from .result_base import (
    SessionBase,
    SuiteBase,
    RunBase,
    RollupBase,
    RetriedBase,
    TestLogBase,
)
from .config_base import ConfigBase, ExportBase, TestConfigBase, MigrationBase
from .static_base import AttachmentBase, StaticBase
from .dynamic_base import TaskBase, PrunedBase
from .attachmentBase import AssertBase

__all__ = [
    SessionBase,
    SuiteBase,
    MigrationBase,
    RunBase,
    TestConfigBase,
    AttachmentBase,
    TaskBase,
    PrunedBase,
    ExportBase,
    ConfigBase,
    RollupBase,
    StaticBase,
    RetriedBase,
    AssertBase,
    TestLogBase,
]



import base64
from sanic.blueprints import Blueprint
from sanic.response import text, HTTPResponse, JSONResponse
from loguru import logger
from sanic.request import Request
from handshake.services.DBService.models.types import (
    AddAttachmentForEntity,
)
from handshake.services.DBService.shared import root_dir
from handshake.services.DBService.models.static_base import (
    StaticBase,
)
from handshake.services.SchedularService.constants import writtenAttachmentFolderName
from handshake.services.Endpoints.blueprints.utils import extractPayload, attachWarn
from handshake.services.Endpoints.define_api import definition

writeServices = Blueprint("WriteService", url_prefix="/write")


@writeServices.on_response
async def handle_response(request: Request, response: JSONResponse):
    if 200 <= response.status < 300:
        return response

    payload = extractPayload(request, response)
    await attachWarn(payload, request.url)
    return JSONResponse(body=payload, status=response.status)


@writeServices.put("/addAttachmentForEntity", error_format="json")
@definition(
    summary="adds an image to the specified entity",
    description="saves an image and then attaches it to a specified entity",
    tag="add",
    deprecated=True,
    body={"application/json": AddAttachmentForEntity},
)
async def saveImage(request: Request) -> HTTPResponse:
    attachment = AddAttachmentForEntity.model_validate(request.json)

    record = await StaticBase.create(
        entity_id=attachment.entityID,
        description=attachment.description,
        type=attachment.type,
    )

    file_name = f"{record.attachmentID}.{record.type.lower()}"
    logger.info("Received a file from user, saving it as {}", file_name)
    root = root_dir() / writtenAttachmentFolderName
    root.mkdir(exist_ok=True)
    test_root = root / str((await (await record.entity).session).test_id)
    test_root.mkdir(exist_ok=True)

    file = test_root / file_name
    file.write_bytes(base64.b64decode(attachment.value))

    await record.update_from_dict(
        dict(
            attachmentValue=dict(value=file_name, title=attachment.title),
        )
    )

    await record.save()
    return text(str(record.attachmentID), status=201)



from handshake.services.DBService.models.dynamic_base import TaskBase
from handshake.services.DBService.models.result_base import SuiteBase, Status, SuiteType
from handshake.services.SchedularService.modifySuites import patchTestSuite
from handshake.services.SchedularService.constants import JobType
from handshake.services.SchedularService.completeTestRun import patchTestRun
from handshake.services.SchedularService.pruneTasks import pruneTasks
from loguru import logger
from tortoise.expressions import Q, Subquery, F
from tortoise.functions import Min
from asyncio import TaskGroup
from handshake.services.SchedularService.register import skip_test_run


async def patch_jobs():
    while True:
        prune_task = await TaskBase.filter(type=JobType.PRUNE_TASKS).first()
        if prune_task:
            await pruneTasks(prune_task.ticketID)

        async with TaskGroup() as patcher:
            # list of tasks which were not picked and processed and are specific to MODIFY_SUITE
            tasks = (
                await TaskBase.filter(
                    Q(picked=False) & Q(processed=False) & Q(type=JobType.MODIFY_SUITE)
                )
                .order_by("dropped")  # old are on top
                .only("ticketID")
                .values_list("ticketID", flat=True)
            )

            if not tasks:
                break

            # providers are list of child suites that needs to be processed and whose parents
            # are planned for processing
            providers = SuiteBase.filter(
                Q(parent__in=tasks)
                & Q(standing=Status.YET_TO_CALCULATE)
                & Q(suiteType=SuiteType.SUITE)
            )
            # these are the list of the suites that needs to be processed and can be processed
            required_suites = SuiteBase.filter(
                Q(suiteID__in=tasks)
                & ~Q(suiteID__in=Subquery(providers.values("parent")))
            )

            # get list of tasks which do not have children that are yet to processed
            to_process = await TaskBase.filter(
                Q(
                    ticketID__in=Subquery(
                        SuiteBase.filter(
                            # these are the list of suites that can be processed
                            # but filters out the suites whose previous retries were not processed
                            Q(suiteID__in=Subquery(required_suites.values("suiteID")))
                            & Q(
                                retried=Subquery(
                                    required_suites.only("retried")
                                    .annotate(minRetries=Min("retried"))
                                    .first()
                                    .values("minRetries")
                                )
                            )
                        ).values("suiteID")
                    )
                )
            ).all()

            # so we process the suites whose child suites are all processed and their previous retries were processed

            # if there are no such tasks to process then check if the tasks are there for the child suites
            if not (
                to_process
                or await TaskBase.filter(
                    Q(ticketID=Subquery(providers.values("suiteID")))
                ).exists()
            ):
                # if so mark each of the tests runs as banned.
                for task in await TaskBase.filter(Q(ticketID__in=tasks)).all():
                    await skip_test_run(
                        task.test_id,
                        "Failed to find a way to process a parent suite, as the child suite was not registered.",
                        parent_suite=task.ticketID,
                        job=task.type,
                    )

            for task in to_process:
                task.picked = True
                await task.save()

                patcher.create_task(
                    patchTestSuite(task.ticketID, task.test_id),
                    name=task.ticketID,
                )

            # NOTE: make sure to pick the task before adding a new task

    logger.debug("Processing Test Runs")
    async with TaskGroup() as patcher:
        for job in await TaskBase.filter(
            Q(type=JobType.MODIFY_TEST_RUN) & Q(picked=False) & Q(processed=False)
        ).all():
            job.picked = True
            await job.save()
            patcher.create_task(patchTestRun(job.ticketID), name=job.ticketID)

    logger.debug("Done!")



from enum import StrEnum
from typing import Tuple


class MigrationStatus(StrEnum):
    PASSED = "PASSED"
    PENDING = "PENDING"
    FAILED = "FAILED"


class MigrationTrigger(StrEnum):
    AUTOMATIC = "AUTO"
    CLI = "CLI"


class Status(StrEnum):
    PASSED = "PASSED"
    PENDING = "PENDING"
    FAILED = "FAILED"
    SKIPPED = "SKIPPED"
    YET_TO_CALCULATE = (
        "YET_TO_CALC"  # needs to be updated by our server to either passed or failed
    )
    # yet_to_calc is mostly seen for the suite
    RETRIED = "RETRIED"


class SuiteType(StrEnum):
    TEST = "TEST"
    SUITE = "SUITE"


class AttachmentType(StrEnum):
    LABEL = "LABEL"
    ENV = "ENV"
    CONFIG = "CONFIG"
    VIDEO = "VIDEO"
    PNG = "PNG"
    DESC = "DESC"
    IMG = "IMAGE"
    LINK = "LINK"
    ASSERT = "ASSERT"


class LogType(StrEnum):
    ERROR = "ERROR"
    WARN = "WARN"
    INFO = "INFO"


class PrunedRecords(StrEnum):
    RUN = "RUN"


class ConfigKeys(StrEnum):
    maxRunsPerProject = "MAX_RUNS_PER_PROJECT"
    version = "VERSION"
    recentlyDeleted = "RECENTLY_DELETED"
    reset_test_run = "RESET_FIX_TEST_RUN"
    py_version = "PY_VERSION"
    last_patch = "LAST_PATCH"
    last_migration = "LAST_MIGRATION"
